﻿using System.Net;
using System.Net.Security;
using System.Net.WebSockets;
using System.Security.Cryptography.X509Certificates;
using Microsoft.AspNetCore.Server.Kestrel.Https;

namespace PCC2.Networking.KestrelSsl;

public delegate X509Certificate2? ServerCertificateSelector(string? sni);

public delegate bool ClientCerValidator(X509Certificate2 clientCert, X509Chain? chain, SslPolicyErrors errors);

public delegate Task WebSocketConnectionHandler(X509Certificate2? clientCert, WebSocket webSocket);

public class KestrelSslListener(
    EndPoint localEndPoint,
    ServerCertificateSelector serverCerSelector,
    ClientCerValidator clientCerValidator,
    WebSocketConnectionHandler connectionHandler,
    ClientCertificateMode clientCertificateMode,
    ILogger<KestrelSslListener> logger)
{
    private IHost? _host;

    public async Task StartAsync()
    {
        if (_host != null) throw new InvalidOperationException("Server is already running.");

        _host = Host.CreateDefaultBuilder()
            .ConfigureLogging(logging =>
            {
                logging.ClearProviders(); // 清除默认日志提供者避免混淆控制台输出
                logging.AddProvider(new LoggerProvider(logger)); // 添加自定义日志提供者
            })
            //.ConfigureLogging(loggingOptions => // 遇到问题可以打开这里查看细节
            //{
            //    loggingOptions.SetMinimumLevel(LogLevel.Trace);
            //})
            .ConfigureWebHostDefaults(webBuilder =>
            {
                webBuilder
                    .UseUrls()
                    .ConfigureKestrel(serverOptions =>
                    {
                        serverOptions.Listen(localEndPoint, listenOptions =>
                        {
                            listenOptions.UseHttps(httpsOptions =>
                            {
                                httpsOptions.ServerCertificateSelector = (_, sni) => serverCerSelector(sni);
                                httpsOptions.ClientCertificateValidation = (clientCert, chain, errors) => clientCerValidator(clientCert, chain, errors);
                                httpsOptions.ClientCertificateMode = clientCertificateMode;
                            });
                        });
                    })
                    .Configure(app =>
                    {
                        app.UseWebSockets();
                        app.Use(async (HttpContext ctx, RequestDelegate _) =>
                        {
                            if (ctx.WebSockets.IsWebSocketRequest == false)
                            {
                                ctx.Response.StatusCode = (int)HttpStatusCode.BadRequest;
                                await ctx.Response.WriteAsync("Bad request, only WebSocket request accept.");
                                return;
                            }

                            var ws = await ctx.WebSockets.AcceptWebSocketAsync();
                            var clientCer = ctx.Connection.ClientCertificate;

                            try
                            {
                                await connectionHandler(clientCer, ws);
                            }
                            catch (Exception ex)
                            {
                                logger.LogError(ex, nameof(connectionHandler));
                            }
                            finally
                            {
                                if (ws.State != WebSocketState.Closed)
                                {
                                    await ws.CloseAsync(WebSocketCloseStatus.InternalServerError, "Internal error", CancellationToken.None);
                                }
                                ws.Dispose(); // 确保资源释放
                            }
                        });
                    });
            })
            .Build();

        await _host.StartAsync();
    }

    public async Task StopAsync()
    {
        if (_host == null)
            throw new InvalidOperationException("Server is not running.");

        await _host.StopAsync();
        logger.LogInformation("Server stopped.");
        _host = null;
    }

    private class LoggerProvider(ILogger logger) : ILoggerProvider
    {
        public ILogger CreateLogger(string categoryName)
        {
            return new CustomLogger(logger, categoryName);
        }

        public void Dispose()
        {
            // 处理必要的清理 这里没有
        }
    }

    private class CustomLogger(ILogger logger, string categoryName) : ILogger
    {
        public IDisposable? BeginScope<TState>(TState state) where TState : notnull => logger.BeginScope(state);

        public bool IsEnabled(LogLevel logLevel) => logger.IsEnabled(logLevel);

        public void Log<TState>(LogLevel logLevel, EventId eventId, TState state, Exception? exception, Func<TState, Exception?, string> formatter)
        {
            if (formatter == null!)
            {
                logger.Log(logLevel, eventId, categoryName + ": " + state, exception);
                return;
            }

            logger.Log(logLevel, eventId, categoryName + ": " + formatter(state, exception), exception);
        }
    }
}
