﻿using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Net.WebSockets;
using System.Security.Cryptography.X509Certificates;

namespace PCC2.Networking;

public class SniWssClient(string sni, string targetHost, int targetPort) : IDisposable
{
    private readonly ClientWebSocket _underlying = new();
    private SimpleHttpProxyServer? _proxy;

    // 封装客户端证书
    public X509CertificateCollection ClientCertificates
    {
        get => _underlying.Options.ClientCertificates;
        set => _underlying.Options.ClientCertificates = value;
    }

    // 封装远程证书验证回调
    public RemoteCertificateValidationCallback? RemoteCertificateValidationCallback
    {
        get => _underlying.Options.RemoteCertificateValidationCallback;
        set => _underlying.Options.RemoteCertificateValidationCallback = value;
    }

    public async Task ConnectAsync(CancellationToken cancellationToken = default)
    {
        // 创建并启动代理
        if (_proxy == null)
        {
            _proxy = new SimpleHttpProxyServer(targetHost, targetPort);
            _proxy.Start();
        }

        // 配置 WebSocket 使用代理
        var proxyEndpoint = _proxy.EndPoint;
        _underlying.Options.Proxy = new WebProxy($"http://{proxyEndpoint.Address}:{proxyEndpoint.Port}");

        // 使用提供的 URI 发起连接
        await _underlying.ConnectAsync(new Uri($"wss://{sni}"), cancellationToken);
    }

    public WebSocketState State => _underlying.State;

    public async Task SendAsync(ArraySegment<byte> buffer, WebSocketMessageType messageType, bool endOfMessage, CancellationToken cancellationToken) => await _underlying.SendAsync(buffer, messageType, endOfMessage, cancellationToken);

    public async Task<WebSocketReceiveResult> ReceiveAsync(ArraySegment<byte> buffer, CancellationToken cancellationToken) => await _underlying.ReceiveAsync(buffer, cancellationToken);

    public async Task CloseAsync(WebSocketCloseStatus closeStatus, string statusDescription, CancellationToken cancellationToken) => await _underlying.CloseAsync(closeStatus, statusDescription, cancellationToken);

    public void Dispose()
    {
        _proxy?.Dispose();
        _underlying.Dispose();
    }

    // 内部 SOCKS 代理类 不适用 准备移除
    private class SocksProxy(string targetHost, int targetPort) : IDisposable
    {
        private readonly TcpListener _listener = new(IPAddress.Loopback, 0);
        private TcpClient? _client;

        public IPEndPoint ProxyEndpoint => (IPEndPoint)_listener.LocalEndpoint;

        public void Start()
        {
            _listener.Start();
            _ = Task.Run(async () =>
            {
                while (true)
                {
                    var client = await _listener.AcceptTcpClientAsync();
                    _client = client;
                    _ = HandleClientAsync(client);
                }
            });
        }

        private async Task HandleClientAsync(TcpClient client)
        {
            await using var clientStream = client.GetStream();

            // SOCKS 协商
            await NegotiateAsync(clientStream);

            // 转发流量
            using var targetClient = new TcpClient();
            await targetClient.ConnectAsync(targetHost, targetPort);
            await using var targetStream = targetClient.GetStream();

            var forward = ForwardAsync(clientStream, targetStream);
            var backward = ForwardAsync(targetStream, clientStream);
            await Task.WhenAll(forward, backward);
        }

        private static async Task NegotiateAsync(NetworkStream stream)
        {
            var buffer = new byte[256];
            var bytesRead = await stream.ReadAsync(buffer, 0, buffer.Length);
            if (bytesRead < 3 || buffer[0] != 0x05)
                throw new InvalidOperationException("Invalid SOCKS version.");

            await stream.WriteAsync(new byte[] { 0x05, 0x00 });
        }

        private static async Task ForwardAsync(Stream input, Stream output)
        {
            var buffer = new byte[8192];
            int bytesRead;
            while ((bytesRead = await input.ReadAsync(buffer, 0, buffer.Length)) > 0)
            {
                await output.WriteAsync(buffer, 0, bytesRead);
            }
        }

        public void Dispose()
        {
            _listener.Stop();
            _client?.Dispose();
        }
    }

    // 内部HTTP代理，用于建立隧道以忽略SNI主机
    public class SimpleHttpProxyServer(string targetHost, int targetPort)
    {
        private TcpListener _listener;

        public IPEndPoint EndPoint => (IPEndPoint)_listener.LocalEndpoint;

        public void Start()
        {
            _listener = new TcpListener(IPAddress.Loopback, 0);
            _listener.Start();

            _ = Task.Run(async () =>
            {
                while (true)
                {
                    var client = await _listener.AcceptTcpClientAsync();
                    _ = Task.Run(async () =>
                    {
                        try
                        {
                            await HandleClientAsync(client);
                        }
                        finally
                        {
                            client.Dispose();
                        }
                    });
                }
            });
        }

        private static readonly byte[] _dummyBuffer = new byte[4096];
        private static readonly ReadOnlyMemory<byte> fixedResponse = new("HTTP/1.1 200 OK\r\n\r\n"u8.ToArray());

        private async Task HandleClientAsync(TcpClient client)
        {
            var inboundStream = client.GetStream();

            // EAT REQUEST assume as CONNECT
            var r = inboundStream.Read(_dummyBuffer, 0, _dummyBuffer.Length);

            // Step 1: Create a TCP connection to the target host (use targetHost to override)
            var targetTcpClient = new TcpClient();
            await targetTcpClient.ConnectAsync(targetHost, targetPort);
            var outboundStream = targetTcpClient.GetStream();

            // Step 2: Send a 200 OK response to the client
            await inboundStream.WriteAsync(fixedResponse, CancellationToken.None);

            // Step 3: Relay data between the client and the target server
            await Task.WhenAny(inboundStream.CopyToAsync(outboundStream), outboundStream.CopyToAsync(inboundStream));
        }

        public void Dispose()
        {
            _listener.Stop();
        }
    }
}
